@@ -22,12 +22,15 @@ from openshift.dynamic.exceptions import NotFoundError, UnprocessibleEntityError
 
 from jinja2 import Environment, FileSystemLoader
 
-from .ocp import getConsoleURL, waitForCRD, waitForDeployment
+from .ocp import getConsoleURL, waitForCRD, waitForDeployment, crdExists
+from .mas import waitForPVC, patchPendingPVC
 
 logger = logging.getLogger(__name__)
 
 
-def installOpenShiftPipelines(dynClient: DynamicClient) -> bool:
+# customStorageClassName is used when no default Storageclass is available on cluster,
+# openshift-pipelines creates PVC which looks for default. customStorageClassName is patched into PVC when default is unavailable.
+def installOpenShiftPipelines(dynClient: DynamicClient, customStorageClassName: str = None) -> bool:
     """
     Install the OpenShift Pipelines Operator and wait for it to be ready to use
     """
@@ -36,25 +39,29 @@ def installOpenShiftPipelines(dynClient: DynamicClient) -> bool:
 
     # Create the Operator Subscription
     try:
-        manifest = packagemanifestAPI.get(name="openshift-pipelines-operator-rh", namespace="openshift-marketplace")
-        defaultChannel = manifest.status.defaultChannel
-        catalogSource = manifest.status.catalogSource
-        catalogSourceNamespace = manifest.status.catalogSourceNamespace
+        if not crdExists(dynClient, "pipelines.tekton.dev"):
+            manifest = packagemanifestAPI.get(name="openshift-pipelines-operator-rh", namespace="openshift-marketplace")
+            defaultChannel = manifest.status.defaultChannel
+            catalogSource = manifest.status.catalogSource
+            catalogSourceNamespace = manifest.status.catalogSourceNamespace
 
-        logger.info(f"OpenShift Pipelines Operator Details: {catalogSourceNamespace}/{catalogSource}@{defaultChannel}")
+            logger.info(f"OpenShift Pipelines Operator Details: {catalogSourceNamespace}/{catalogSource}@{defaultChannel}")
 
-        templateDir = path.join(path.abspath(path.dirname(__file__)), "templates")
-        env = Environment(
-            loader=FileSystemLoader(searchpath=templateDir)
-        )
-        template = env.get_template("subscription.yml.j2")
-        renderedTemplate = template.render(
-            pipelines_channel=defaultChannel,
-            pipelines_source=catalogSource,
-            pipelines_source_namespace=catalogSourceNamespace
-        )
-        subscription = yaml.safe_load(renderedTemplate)
-        subscriptionsAPI.apply(body=subscription, namespace="openshift-operators")
+            templateDir = path.join(path.abspath(path.dirname(__file__)), "templates")
+            env = Environment(
+                loader=FileSystemLoader(searchpath=templateDir)
+            )
+            template = env.get_template("subscription.yml.j2")
+            renderedTemplate = template.render(
+                subscription_name="openshift-pipelines-operator",
+                subscription_namespace="openshift-operators",
+                package_name="openshift-pipelines-operator-rh",
+                package_channel=defaultChannel,
+                catalog_name=catalogSource,
+                catalog_namespace=catalogSourceNamespace
+            )
+            subscription = yaml.safe_load(renderedTemplate)
+            subscriptionsAPI.apply(body=subscription, namespace="openshift-operators")
 
     except NotFoundError:
         logger.warning("Error: Couldn't find package manifest for Red Hat Openshift Pipelines Operator")
@@ -75,11 +82,31 @@ def installOpenShiftPipelines(dynClient: DynamicClient) -> bool:
     foundReadyWebhook = waitForDeployment(dynClient, namespace="openshift-pipelines", deploymentName="tekton-pipelines-webhook")
     if foundReadyWebhook:
         logger.info("OpenShift Pipelines Webhook is installed and ready")
-        return True
     else:
         logger.error("OpenShift Pipelines Webhook is NOT installed and ready")
         return False
 
+    # Wait for the postgredb-tekton-results-postgres-0 PVC to be ready
+    # this PVC doesn't come up when there's no default storage class is in the cluster,
+    # this is causing the pvc to be in pending state and causing the tekton-results-postgres statefulSet in pending,
+    # due to these resources not coming up, the MAS pre-install check in the pipeline times out checking the health of this statefulSet,
+    # causing failure in pipeline.
+    # Refer https://github.com/ibm-mas/cli/issues/1511
+    logger.debug("Waiting for postgredb-tekton-results-postgres-0 PVC to be ready")
+    foundReadyPVC = waitForPVC(dynClient, namespace="openshift-pipelines", pvcName="postgredb-tekton-results-postgres-0")
+    if foundReadyPVC:
+        logger.info("OpenShift Pipelines postgres is installed and ready")
+        return True
+    else:
+        patchedPVC = patchPendingPVC(dynClient, namespace="openshift-pipelines", pvcName="postgredb-tekton-results-postgres-0", storageClassName=customStorageClassName)
+        if patchedPVC:
+            logger.info("OpenShift Pipelines postgres is installed and ready")
+            return True
+        else:
+            logger.error("OpenShift Pipelines postgres PVC is NOT ready")
+            return False
+
+
 def updateTektonDefinitions(namespace: str, yamlFile: str) -> None:
     """
     Install/update the MAS tekton pipeline and task definitions
@@ -96,25 +123,26 @@ def updateTektonDefinitions(namespace: str, yamlFile: str) -> None:
     for line in result.split("\n"):
         logger.debug(line)
 
-def preparePipelinesNamespace(dynClient: DynamicClient, instanceId: str=None, storageClass: str=None, accessMode: str=None, waitForBind: bool=True):
+
+def preparePipelinesNamespace(dynClient: DynamicClient, instanceId: str = None, storageClass: str = None, accessMode: str = None, waitForBind: bool = True, configureRBAC: bool = True):
     templateDir = path.join(path.abspath(path.dirname(__file__)), "templates")
     env = Environment(
         loader=FileSystemLoader(searchpath=templateDir)
     )
-
     if instanceId is None:
-        namespace = f"mas-pipelines"
+        namespace = "mas-pipelines"
         template = env.get_template("pipelines-rbac-cluster.yml.j2")
     else:
         namespace = f"mas-{instanceId}-pipelines"
         template = env.get_template("pipelines-rbac.yml.j2")
 
-    # Create RBAC
-    renderedTemplate = template.render(mas_instance_id=instanceId)
-    logger.debug(renderedTemplate)
-    crb = yaml.safe_load(renderedTemplate)
-    clusterRoleBindingAPI = dynClient.resources.get(api_version="rbac.authorization.k8s.io/v1", kind="ClusterRoleBinding")
-    clusterRoleBindingAPI.apply(body=crb, namespace=namespace)
+    if configureRBAC:
+        # Create RBAC
+        renderedTemplate = template.render(mas_instance_id=instanceId)
+        logger.debug(renderedTemplate)
+        crb = yaml.safe_load(renderedTemplate)
+        clusterRoleBindingAPI = dynClient.resources.get(api_version="rbac.authorization.k8s.io/v1", kind="ClusterRoleBinding")
+        clusterRoleBindingAPI.apply(body=crb, namespace=namespace)
 
     # Create PVC (instanceId namespace only)
     if instanceId is not None:
@@ -141,8 +169,9 @@ def preparePipelinesNamespace(dynClient: DynamicClient, instanceId: str=None, st
                 logger.debug(configPVC)
                 sleep(15)
 
-def prepareInstallSecrets(dynClient: DynamicClient, instanceId: str, slsLicenseFile: str, additionalConfigs: dict=None, certs: str=None, podTemplates: str=None) -> None:
-    namespace=f"mas-{instanceId}-pipelines"
+
+def prepareInstallSecrets(dynClient: DynamicClient, instanceId: str, slsLicenseFile: str = None, additionalConfigs: dict = None, certs: str = None, podTemplates: str = None) -> None:
+    namespace = f"mas-{instanceId}-pipelines"
     secretsAPI = dynClient.resources.get(api_version="v1", kind="Secret")
 
     # 1. Secret/pipeline-additional-configs
@@ -171,10 +200,16 @@ def prepareInstallSecrets(dynClient: DynamicClient, instanceId: str, slsLicenseF
     except NotFoundError:
         pass
 
-    # TODO: Convert this to using secretsAPI.create()
-    result = kubectl.run(subcmd_args=['-n', namespace, 'create', 'secret', 'generic', 'pipeline-sls-entitlement', '--from-file', slsLicenseFile])
-    for line in result.split("\n"):
-        logger.debug(line)
+    if slsLicenseFile is None:
+        slsLicenseFile = {
+            "apiVersion": "v1",
+            "kind": "Secret",
+            "type": "Opaque",
+            "metadata": {
+                "name": "pipeline-sls-entitlement"
+            }
+        }
+    secretsAPI.create(body=slsLicenseFile, namespace=namespace)
 
     # 3. Secret/pipeline-certificates
     # -------------------------------------------------------------------------
@@ -214,6 +249,7 @@ def prepareInstallSecrets(dynClient: DynamicClient, instanceId: str, slsLicenseF
         }
     secretsAPI.create(body=podTemplates, namespace=namespace)
 
+
 def testCLI() -> None:
     pass
     # echo -n "Testing availability of $CLI_IMAGE in cluster ..."
@@ -246,10 +282,12 @@ def testCLI() -> None:
     #     exit 1
     # fi
 
+
 def launchUpgradePipeline(dynClient: DynamicClient,
                           instanceId: str,
                           skipPreCheck: bool = False,
-                          masChannel: str = "") -> str:
+                          masChannel: str = "",
+                          params: dict = {}) -> str:
     """
     Create a PipelineRun to upgrade the chosen MAS instance
     """
@@ -266,7 +304,8 @@ def launchUpgradePipeline(dynClient: DynamicClient,
         timestamp=timestamp,
         mas_instance_id=instanceId,
         skip_pre_check=skipPreCheck,
-        mas_channel=masChannel
+        mas_channel=masChannel,
+        **params
     )
     logger.debug(renderedTemplate)
     pipelineRun = yaml.safe_load(renderedTemplate)
@@ -275,8 +314,10 @@ def launchUpgradePipeline(dynClient: DynamicClient,
     pipelineURL = f"{getConsoleURL(dynClient)}/k8s/ns/mas-{instanceId}-pipelines/tekton.dev~v1beta1~PipelineRun/{instanceId}-upgrade-{timestamp}"
     return pipelineURL
 
+
 def launchUninstallPipeline(dynClient: DynamicClient,
                             instanceId: str,
+                            droNamespace: str,
                             certManagerProvider: str = "redhat",
                             uninstallCertManager: bool = False,
                             uninstallGrafana: bool = False,
@@ -317,7 +358,8 @@ def launchUninstallPipeline(dynClient: DynamicClient,
         ibm_catalogs_action=ibmCatalogAction,
         mongodb_action=mongoDbAction,
         sls_action=slsAction,
-        uds_action=udsAction
+        uds_action=udsAction,
+        dro_namespace=droNamespace
     )
     logger.debug(renderedTemplate)
     pipelineRun = yaml.safe_load(renderedTemplate)
@@ -326,6 +368,7 @@ def launchUninstallPipeline(dynClient: DynamicClient,
     pipelineURL = f"{getConsoleURL(dynClient)}/k8s/ns/mas-{instanceId}-pipelines/tekton.dev~v1beta1~PipelineRun/{instanceId}-uninstall-{timestamp}"
     return pipelineURL
 
+
 def launchPipelineRun(dynClient: DynamicClient, namespace: str, templateName: str, params: dict) -> str:
     pipelineRunsAPI = dynClient.resources.get(api_version="tekton.dev/v1beta1", kind="PipelineRun")
     timestamp = datetime.now().strftime("%y%m%d-%H%M")
@@ -346,6 +389,7 @@ def launchPipelineRun(dynClient: DynamicClient, namespace: str, templateName: st
     pipelineRunsAPI.apply(body=pipelineRun, namespace=namespace)
     return timestamp
 
+
 def launchInstallPipeline(dynClient: DynamicClient, params: dict) -> str:
     """
     Create a PipelineRun to install the chosen MAS instance (and selected dependencies)
@@ -357,11 +401,24 @@ def launchInstallPipeline(dynClient: DynamicClient, params: dict) -> str:
     pipelineURL = f"{getConsoleURL(dynClient)}/k8s/ns/mas-{instanceId}-pipelines/tekton.dev~v1beta1~PipelineRun/{instanceId}-install-{timestamp}"
     return pipelineURL
 
+
+def launchInstallPipelineForAiservice(dynClient: DynamicClient, params: dict) -> str:
+    """
+    Create a PipelineRun to install the Aiservice
+    """
+    instanceId = params["aibroker_instance_id"]
+    namespace = f"mas-{instanceId}-pipelines"
+    timestamp = launchPipelineRun(dynClient, namespace, "pipelinerun-aiservice-install", params)
+
+    pipelineURL = f"{getConsoleURL(dynClient)}/k8s/ns/mas-{instanceId}-pipelines/tekton.dev~v1beta1~PipelineRun/{instanceId}-install-{timestamp}"
+    return pipelineURL
+
+
 def launchUpdatePipeline(dynClient: DynamicClient, params: dict) -> str:
     """
     Create a PipelineRun to update the Maximo Operator Catalog
     """
-    namespace = f"mas-pipelines"
+    namespace = "mas-pipelines"
     timestamp = launchPipelineRun(dynClient, namespace, "pipelinerun-update", params)
 
     pipelineURL = f"{getConsoleURL(dynClient)}/k8s/ns/mas-pipelines/tekton.dev~v1beta1~PipelineRun/mas-update-{timestamp}"
